module AASM
  module Persistence
    module NContentRestPersistence
      # This method:
      #
      # * extends the model with ClassMethods
      # * includes InstanceMethods
      #
      # Adds
      #
      # before_validation :aasm_ensure_initial_state, :on => :create
      #
      # As a result, it doesn't matter when you define your methods - the following 2 are equivalent
      #
      # class Foo < ActiveRecord::Base
      # def aasm_write_state(state)
      # "bar"
      # end
      # include AASM
      # end
      #
      # class Foo < ActiveRecord::Base
      # include AASM
      # def aasm_write_state(state)
      # "bar"
      # end
      # end
      #
      def self.included(base)
        base.send(:include, AASM::Persistence::Base)
        base.extend AASM::Persistence::NContentRestPersistence::ClassMethods
        base.send(:include, AASM::Persistence::NContentRestPersistence::InstanceMethods)

        base.before_validation(:aasm_ensure_initial_state, :on => :create)

        # ensure initial aasm state even when validations are skipped
        base.before_create(:aasm_ensure_initial_state)

        # ensure state is in the list of states
        # base.validate :aasm_validate_states
      end

      module ClassMethods

        # def find_in_state(number, state, *args)
        #   with_state_scope state do
        #     find(number, *args)
        #   end
        # end

        # def count_in_state(state, *args)
        #   with_state_scope state do
        #     count(*args)
        #   end
        # end

        # def calculate_in_state(state, *args)
        #   with_state_scope state do
        #     calculate(*args)
        #   end
        # end

        # protected
        # def with_state_scope(state)
        #   with_scope :find => {:conditions => ["#{table_name}.#{aasm.attribute_name} = ?", state.to_s]} do
        #     yield if block_given?
        #   end
        # end
      end

      module InstanceMethods

        # Writes <tt>state</tt> to the state column and persists it to the server
        #
        # foo = Foo.find(1)
        # foo.aasm.current_state # => :opened
        # foo.close!
        # foo.aasm.current_state # => :closed
        # Foo.find(1).aasm.current_state # => :closed
        #
        # NOTE: intended to be called from an event
        def aasm_write_state(new_state)

          previous_state = self.public_send self.class.aasm.attribute_name
          self.public_send "#{self.class.aasm.attribute_name}=", new_state

          success = begin
            data = {
              name: aasm.current_event.to_s.tr('!',''),
              target_type: self.class.name.demodulize,
              target_id: self.id
            }

            unless NContent::SDK.config.state_event_delay < 1
              msg = "Will wait #{NContent::SDK.config.state_event_delay} seconds"
              NContent::SDK.logger.debug "#{msg} before posting the event..."
              sleep NContent::SDK.config.state_event_delay
            end

            response = NContent::SDK::RESTClient.post "/events.json" do |req|
              req.headers[:content_type] = 'application/json'
              req.body = ActiveSupport::JSON.encode(state_event: data)
            end

            reload

            # Return true if the actual state is the same as the expected state:
            self.public_send(self.class.aasm.attribute_name) == new_state.to_s
          end

          unless success
            self.public_send "#{self.class.aasm.attribute_name}=", previous_state
            return false
          end

          true
        end

        def aasm_write_state_without_persistence(new_state)
          self.public_send(self.class.aasm.attribute_name) == new_state.to_s
          true
        end

      private

        def aasm_fire_event(name, options, *args, &block)
          success = super

          if success
            if options[:persist]
              event = self.class.aasm.state_machine.events[name]
              event.fire_callbacks(:after_commit, self, *args)
            else
              # TODO: Implement a "state change queue" to make .save send the state changes
              # to the server when no bang (options[:persist] == false) was given...
            end
          end

          success
        end

        def aasm_ensure_initial_state
          send("#{self.class.aasm.attribute_name}=", aasm.enter_initial_state.to_s) \
            if send(self.class.aasm.attribute_name).blank?
        end

      end

    end
  end
end
