module NaranyaEcm::Behaviors
  module MediaMethods

    extend ActiveSupport::Concern

    def find_media_by_role(role_name)
      mr = self.media_resources.nil? ? [] : self.media_resources
      mr.detect { |mr| mr.roles.include? role_name }
    end

    def find_media_url_by_role(role_name)
      mr = self.find_media_by_role role_name
      mr.present? ? mr.downloadable_url : nil
    end

    def select_media_by_role(role_name)
      self.media_resources.select { |mr| mr.roles.include? role_name }
    end

    def select_media_urls_by_role(role_name)
      mr = self.select_media_by_role(role_name).map(&:downloadable_url)
      mr.flatten.uniq
    end

    # Returns an Array of MediaResource objects which match the given options in their role list.
    #
    # ==== Attributes
    #
    # * +options+ - The given options to select the MediaResource objects
    #
    # ==== Options
    #
    # You may which to break out options as a separate item since there maybe
    # multiple items. Note options are prefixed with a colon, denoting them
    # as a
    #
    # * +:match+ - An Array of roles that *must* exist in matching MediaResource objects.
    # * +:prefer+ - An Array of values consisting of:
    #   * A String indicating a desired role
    #   * A Hash containing the desired role as key, and an Array of fallback roles as value.
    #
    # ==== Examples
    #
    # Illustrate the behaviour of the method using examples. Indent examples:
    #
    #   base = Base.new("Example String")
    #   base.method_name("Example", "more")
    def select_media_by_roles(options)

      options = options.with_indifferent_access if options.is_a?(Hash) && !options.is_a?(ActiveSupport::HashWithIndifferentAccess)

      match = options.is_a?(Hash) ? options[:match] : options
      match = [match.to_s] if match.is_a?(String) || match.is_a?(Symbol)
      
      raise ArgumentError, "No match options given" if match.blank?

      except = options.is_a?(Hash) ? options[:except] : []
      except = [except.to_s] if except.is_a?(String) || except.is_a?(Symbol)
      except = [] if except.blank?

      filtered_media = self.media_resources.select do |mr|
        ((mr.roles & match).count == match.count) && ((mr.roles & except).count < 1)
      end

      prefer = options.is_a?(Hash) ? options[:prefer] : nil
      prefer = [prefer.to_s] if prefer.is_a?(String) || prefer.is_a?(Symbol)
      
      filtered_media
    end

    def detect_media_by_roles(options)
      select_media_by_roles(options).first
    end

    def select_media_urls_by_roles(options)
      selected_media = select_media_by_roles(options)
      selected_media.present? ? selected_media.map(&:downloadable_url) : nil
    end

    def detect_media_url_by_roles(options)
      select_media_urls_by_roles(options).first
    end

  end
end