require 'aasm'
require 'aasm/persistence/base'
require 'aasm/persistence/ncontent_rest_persistence'

module NaranyaEcm

  ##
  # Represents a Content unit in the Naranya CaaS Service
  #
  # A content unit is made of by a list of some required attributes,
  # along with any non-required attributes.
  #
  # == Attributes List
  #
  # === Required Attributes
  #
  # In order to create a Content Unit, the Content Object should contain at least the following attributes:
  # * <tt>lifecycle_name:</tt> Indicates whether the Content object follows a <tt>simple</tt> lifecycle, a <tt>validatable</tt> lifecycle (can't publish without some sort of editorial validation first), or any other significant lifecycle.
  # * <tt>title_translations:</tt> At least one localized title is required. This attribute is a Hash-like object, requiring the locale name as a key, and the localized string as value.
  # * <tt>author:</tt> This string represents the name of the author of the content.
  #
  # === Other significant attributes
  #
  # The Naranya CaaS Service is prepared to manage and gives meaning to the following attributes:
  # * <tt>type:</tt> The type of content. Aids the client application in determining how to display and manage the content.
  # * <tt>description_translations:</tt> Localizable descriptions of the Content.
  # * <tt>content_rating:</tt>
  # * <tt>keywords:</tt> Array containing a list of keywords
  # * <tt>main_url:</tt> The main URL pointing to the content AS DISPLAYED BY THE CLIENT APP.
  #
  # == Associations
  #
  # * <tt>category:</tt>
  # * <tt>versions:</tt>
  # * <tt>current_version:</tt>
  # * <tt>latest_version:</tt>
  # * <tt>media_resources:</tt>
  class Content

    # Include the REST persistance methods:
    include NaranyaEcm.document_module

    # Include the timestamping methods:
    include NaranyaEcm::Behaviors::Timestampable

    # Include the search methods:
    include NaranyaEcm::Search::Methods

    # Include the media list methods:
    include NaranyaEcm::Behaviors::MediaMethods

    ##
    # :attr: type
    #
    # Describes the main type of the content - App, Note, etc.
    field :type, type: String

    ##
    # :attr: lifecycle_name
    #
    # Determines the lifecycle for the specific content:
    # * <tt>simple:</tt>       draft -> published -> deactivated
    # * <tt>validatable:</tt>  draft -> awaiting_acceptance -> accepted -> published -> deactivated
    # * <tt>expirable:</tt>    draft -> published -> expired/inactive
    field :lifecycle_name, type: String, default: -> { 'simple' }
    validates :lifecycle_name, presence: true

    ##
    # :attr: lifecycle_state
    #
    # A managed string that describes the current state of this particular content.
    # Depending on the Content's lifecycle, it changes upon several events.
    # * <tt>draft:</tt>
    # * <tt>awaiting_acceptance:</tt>
    # * <tt>accepted:</tt>
    # * <tt>rejected:</tt>
    # * <tt>published:</tt>
    # * <tt>deactivated:</tt>
    field :lifecycle_state, type: String, default: -> { :draft }
    validates :lifecycle_state, presence: true

    ##
    # :attr: title
    #
    # The content title that can be displayed as header, result, etc.
    field :title, type: String, localize: true
    validate :at_least_one_title_translation_must_exist

    ##
    # :attr: title_translations
    #
    # Hash-like containing all the title translations.

    ##
    # :attr: description
    #
    # The localized content description that can be displayed as header, result, etc.
    field :description, type: String, localize: true

    ##
    # :attr: description_translations
    #
    # Hash-like containing all the description translations.

    ##
    # :attr: content_rating
    #
    # Hash-like object containing all the localized content ratings, with the rating system as a Key, and the corresponding ratin object as value.
    field :content_rating, type: Hash

    ##
    # :attr: keywords
    #
    # Array containing a list of keywords
    field :keywords, type: Array

    ##
    # :attr: author
    #
    # Author:
    # The name of the user who created this content.
    field :author, type: String
    validates :author, presence: true

    ##
    # :attr: main_url
    #
    # Main URL:
    # The URL that should be used to render or display the content.
    field :main_url, type: String
    validates :main_url, format: { with: URI.regexp }, allow_nil: true, allow_blank: true

    ##
    # :attr: category
    # :category: Associations
    #
    # The category to which the Content belongs.
    belongs_to :category

    ##
    # :attr: versions
    # :category: Associations
    #
    # A list of ContentVersion objects corresponding to changes in a versionable content.
    has_many :versions, class_name: :ContentVersion

    ##
    # :attr: current_version
    # :category: Associations
    #
    # The ContentVersion object marked as current. It must be published.
    belongs_to :current_version, class_name: :ContentVersion

    ##
    # :attr: latest_version
    # :category: Associations
    #
    # The ContentVersion object marked as latest.
    belongs_to :latest_version, class_name: :ContentVersion

    ##
    # :attr: media_resources
    # :category: Associations
    #
    # A list of MediaResource objects associated directly to the Content.
    has_many :media_resources, class_name: :MediaResource, inverse_of: :content

    ###############################################################
    # StateMachine:
    include AASM
    include AASM::Persistence::NContentRestPersistence

    ##
    # :method: request_an_acceptance_review
    # :call-seq:
    #   request_an_acceptance_review => true or false
    #
    # Triggers the <tt>request_an_acceptance_review</tt> event, and starts the Content transition to an :awaiting_acceptance state.

    ##
    # :method: request_an_acceptance_review!
    # :call-seq:
    #   request_an_acceptance_review => true or Error
    #
    # Same as #request_an_acceptance_review, but raises an error if the transition halted.

    ##
    # :method: accept
    # :call-seq:
    #   accept => true or false
    #
    # Triggers the <tt>accept</tt> event.

    ##
    # :method: accept!
    # :call-seq:
    #   accept! => true or Error
    #
    # Same as #accept, but raises an error if the transition halted.

    ##
    # :method: reject
    # :call-seq:
    #   reject => true or false
    #
    # Triggers the <tt>reject</tt> event, which starts the Content transition to an <tt>:rejected</tt> state.

    ##
    # :method: reject!
    # :call-seq:
    #   reject! => true or Error
    #
    # Same as #reject, but raises an error if the transition did not complete.

    ##
    # :method: publish
    # :call-seq:
    #   publish => true or false
    #
    # Triggers the <tt>publish</tt> event, which starts the Content transition to a <tt>:published</tt> state.

    ##
    # :method: publish!
    # :call-seq:
    #   publish! => true or Error
    #
    # Same as #publish, but raises an error if the transition did not complete.

    ##
    # :method: deactivate
    # :call-seq:
    #   deactivate => true or false
    #
    # Triggers the <tt>deactivate</tt> event, which starts the Content transition to a <tt>:deactivated</tt> state.

    ##
    # :method: deactivate!
    # :call-seq:
    #   deactivate! => true or Error
    #
    # Same as #deactivate, but raises an error if the transition did not complete.

    aasm column: :lifecycle_state do

      state :draft, initial: true
      state :awaiting_acceptance
      state :accepted
      state :rejected
      state :published
      state :deactivated

      #########################################
      # Eventos:

      event :request_an_acceptance_review, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :draft a :awaiting_acceptance,
        transitions from: :draft, to: :awaiting_acceptance, guard: :validatable_lifecycle?
      end

      event :accept, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :awaiting_acceptance a :accepted:
        transitions from: :awaiting_acceptance, to: :accepted
      end

      event :reject, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :awaiting_acceptance a :rejected:
        transitions from: :awaiting_acceptance, to: :rejected
      end

      event :publish, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :draft a :published,
        # si el lifecycle es sencillo:
        transitions from: :draft, to: :published,     guard: :simple_lifecycle?
        transitions from: :validated, to: :published, guard: :validatable_lifecycle?
        transitions from: :deactivated, to: :published, guard: :validatable_lifecycle?
      end

      event :deactivate, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :published a :deactivated:
        transitions from: :published, to: :deactivated
      end
    end

    ##
    # :method: simple_lifecycle?
    # :call-seq:
    #   simple_lifecycle? => true or false
    #
    # Shows whether the Content object has a simple lifecycle.
    def simple_lifecycle?
      self.lifecycle_name == 'simple'
    end

    ##
    # :method: validatable_lifecycle?
    # :call-seq:
    #   validatable_lifecycle? => true or false
    #
    # Shows whether the Content object has a validatable lifecycle.
    def validatable_lifecycle?
      self.lifecycle_name == 'validatable'
    end

    private

      # Validates that at least one title translation exists.
      def at_least_one_title_translation_must_exist
        unless title_translations.present? && title_translations.keys.any?
          errors.add :title_translations, "must have at least one translation"
        end
      end

  end
end
