require 'aasm'
require 'aasm/persistence/base'
require 'aasm/persistence/ncontent_rest_persistence'

module NaranyaEcm
  class ContentVersion

    include NaranyaEcm.document_module
    include NaranyaEcm::Behaviors::Timestampable
    include NaranyaEcm::Behaviors::MediaMethods

    # Description (localized):
    # A localized string that describes an overview of the significant changes
    # particular in this version. Used to list the most significant updates, fixes, etc.
    # to the end user.
    field :description, type: String, localize: true

    # tomar lifecycle de content:
    attr_accessor :lifecycle_name
    #field :lifecycle_name, type: String

    # Version Lifecycle State:
    # A managed string that describes the current state of this particular version.
    # Depending on the Content's lifecycle, it changes upon several events.
    # - draft:
    # - awaiting_acceptance:
    # - accepted:
    # - rejected:
    # - published:
    # - superceded:
    # - deactivated:
    field :lifecycle_state, type: String, default: -> { :draft }
    validates :lifecycle_state, presence: true

    # Version Code:
    # An internal version number. This number is used only to determine whether one version is more recent than another, with higher numbers indicating more recent versions. This is not the version number shown to users; that number is set by the versionName attribute.
    # The value must be set as an integer, such as "100". You can define it however you want, as long as each successive version has a higher number. For example, it could be a build number. Or you could translate a version number in "x.y" format to an integer by encoding the "x" and "y" separately in the lower and upper 16 bits. Or you could simply increase the number by one each time a new version is released.
    field :code, type: Integer
    validates :code, presence: true, format: { with: /[0-9]+/ }

    # Version Name:
    # The version number shown to users.
    # The string has no other purpose than to be displayed to users. The Version Code
    # attribute holds the significant version number used internally.
    field :name, type: String

    # Content
    # The content to which this particular version belongs:
    belongs_to :content, class_name: :Content, inverse_of: :versions

    # MediaResources:
    # A list of media - images, audio, video, etc. associated
    # directly to this particular content version. Some examples:
    # - Icons, Banners, screenshots, gallery images
    has_many :media_resources, class_name: :MediaResource, inverse_of: :content_version

    ###############################################################
    # StateMachine:
    include AASM
    include AASM::Persistence::NContentRestPersistence

    aasm column: :lifecycle_state do

      state :draft, initial: true
      state :awaiting_acceptance
      state :accepted
      state :rejected
      state :published
      state :superseded
      state :deactivated

      #########################################
      # Eventos:

      event :request_an_acceptance_review, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :draft a :awaiting_acceptance,
        transitions from: :draft, to: :awaiting_acceptance, guard: :validatable_lifecycle?
      end

      event :accept, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :awaiting_acceptance a :accepted:
        transitions from: :awaiting_acceptance, to: :accepted
      end

      event :reject, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :awaiting_acceptance a :rejected:
        transitions from: :awaiting_acceptance, to: :rejected
      end

      event :publish, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :draft a :published,
        # si el lifecycle es sencillo:
        transitions from: :draft,       to: :published, guard: :simple_lifecycle?
        transitions from: :accepted,    to: :published, guard: :validatable_lifecycle?
        transitions from: :deactivated, to: :published
      end

      event :supersede, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :published a :superseded:
        transitions from: :published, to: :superseded
      end

      event :deactivate, before: :start_notify_state_change!, after_commit: :finish_notify_state_change! do
        # Iniciar la transición de :published a :deactivated:
        transitions from: :published, to: :deactivated
      end
    end

    def simple_lifecycle?; self.lifecycle_name == 'simple'; end
    def validatable_lifecycle?; self.lifecycle_name == 'validatable'; end

  end
end
