
module NaranyaEcm
  class MediaResource

    include NaranyaEcm.document_module
    include NaranyaEcm::Behaviors::Timestampable

    # Media MIME/Internet Type:
    field :type, type: String

    # The roles of the media file (screenshot, icon, main_downloadable):
    field :roles, type: Array, default: -> {[]}

    # The URL to the downloadable file:
    field :downloadable_url,  type: String
    validates :downloadable_url, presence: true, format: { with: URI.regexp }

    # The access type of the media file (private|public)
    field :access_type, type: String, default: -> { 'public' }
    validates :access_type, presence: true

    # The list of compatible device ids:
    field :compatible_device_ids, type: Array, default: -> { [] }

    # The list of compatible device operating systems:
    field :compatible_device_oses, type: Hash, default: -> { {} }

    ##
    # :attr: media_resources
    # :category: Associations
    #
    # A list of child/processed MediaResource objects associated directly to this MediaResource.
    has_many :media_resources, class_name: :MediaResource, inverse_of: :media_resource

    #belongs_to :parent_document
    field :parent_document_id,   type: String
    field :parent_document_type, type: String
    def as_notification_payload
      instance_payload = super.merge(roles: self.roles)

      if self.parent_document_id.present? && self.parent_document_type.present?
        instance_payload[:parent_id]    = self.parent_document_id
        instance_payload[:parent_class] = "NaranyaEcm::#{self.parent_document_type}"
      end

      instance_payload
    end

    # Returns true if the Media Resource is private
    def private?
      self.access_type == 'private'
    end

    # Returns true if the Media Resource is private
    def public?
      self.access_type == 'public'
    end

    # Returns the downloadable resource key, which depending of the storage engine, is one of the following:
    # - Amazon AWS S3: The downloadable object key
    def downloadable_resource_key
      @downloadable_file_key ||= self.downloadable_path[1..-1] if self.downloadable_path.present?
    end

    # Returns the downloadable resource, which depending of the storage engine, is one of the following:
    # - Amazon AWS S3: A Fog::Storage::AWS::File object, representing the file object in Amazon
    def downloadable_resource
      root_directory.files.head downloadable_resource_key
    end

    def downloadable_uri
      @downloadable_uri ||= if self.downloadable_url.present?
        begin
          URI(self.downloadable_url)
        rescue URI::InvalidURIError
          # Maybe a special character prevents the URI to form...
          URI(URI::encode self.downloadable_url)
        end
      end
    end

    def downloadable_path
      @downloadable_path ||= self.downloadable_uri.path if self.downloadable_uri.present?
    end

    def downloadable_host
      @downloadable_host ||= self.downloadable_uri.host if self.downloadable_uri.present?
    end

    def downloadable_cdn_url(options = {})
      options = options.with_indifferent_access
      ret = self.downloadable_uri.clone

      ret.scheme = options[:scheme] if options.has_key? :scheme
      ret.host   = options[:host]   if options.has_key? :host

      ret.to_s
    end

    # Return the URL used to access the media resource file.
    # - In case of a private resource file, the URL includes parameters used to access the file
    #   for the following 5 minutes, unless an expiration time is given.
    def downloadable_authorized_url(expiration_time = nil)
      if self.public?
        downloadable_resource.public_url
      else
        expiration_time = 5.minutes.from_now unless expiration_time.is_a? Time
        downloadable_resource.url(expiration_time)
      end
    end

    delegate :root_directory, to: :class

    # Returns the storage root directory, depending of the storage engine:
    # - Amazon AWS S3: A Fog::Storage::AWS::Directory object
    def self.root_directory
      NContent::SDK.root_directory
    end

    def move_downloadable_to(path, force_move=false)

      raise ArgumentError, "Given path is empty" unless path.present?
      path = path[1..-1] if path.present? && path[0] == "/"
      raise ArgumentError, "Can't use root (/) as target path" unless path.present?

      path_exists = root_directory.files.head(path).present?

      if !path_exists || (path_exists && force_move)
        old_file = root_directory.files.get downloadable_resource_key

        new_file = root_directory.files.create(
          key: path,
          body: old_file.body,
          content_type: old_file.content_type,
          last_modified: old_file.last_modified,
          :public => public?
        )

        self.downloadable_url = public? ? new_file.public_url : new_file.url(1.minutes.from_now).split("?").first

        old_file.destroy if self.save
        true
      else
        false
      end

    end

    # Generates a MediaProcess with self as the source MediaResource.
    def process_with_profile(given_profile, given_options = {})
      raise ArgumentError, 'Given profile is not a MediaProfile' unless given_profile.is_a? NaranyaEcm::MediaProfile

      # TODO: Eliminar template_id...
      options = given_options.merge media_resource: self, template_id: given_profile.template_id

      NaranyaEcm::MediaProcess.create options
    end

  end
end
