module NaranyaEcm::Rest

  module Associations
    extend ActiveSupport::Concern

    class Association
      attr_reader :klass, :name, :associated_class, :options

      def resolve_class_name(given_class_name)
        given_class_name = "#{@klass.name.deconstantize}::#{given_class_name}" if given_class_name.is_a?(Symbol)
        given_class_name.constantize
      end

      def initialize(klass, name, options = {})
        @klass, @name = klass, name

        @associated_class = if (given_class_name = options.delete(:class_name))
          resolve_class_name(given_class_name)
        else
          resolve_class_name(name.to_s.classify.to_sym)
        end

        @options = options.with_indifferent_access

        klass.send :attr_accessor, name

      end
    end

    class CollectionAssociation < Association
    end

    class SingularAssociation < Association
    end

    class HasMany < Association
      def foreign_key
        @foreign_key ||= @options[:foreign_key] || @options.has_key?(:inverse_of) ? "#{@options[:inverse_of]}_id".to_sym : "#{@klass.name.demodulize.underscore}_id".to_sym
      end
      def initialize(klass, name, options = {})
        super
        # Define the Getter:
        klass.send :define_method, name do
          value = self.instance_variable_get "@#{name}".to_sym
          if value.nil?
            reflection = self.class.reflect_on_association name
            foreign_key = (self.class.name.demodulize.underscore + '_id').to_sym
            value = self.instance_variable_set(
              "@#{name}".to_sym,
              # reflection.associated_class.where(foreign_key => self.id)
              AssociationRelation.new(reflection.associated_class, reflection.foreign_key, self.id)
            )
          end
          value
        end

        # Define the setter
        klass.send :define_method, "#{name}=".to_sym do |new_value|
          reflection = self.class.reflect_on_association(name)
          relation = AssociationRelation.new(reflection.associated_class, reflection.foreign_key, self.id)

          relation.load new_value unless new_value.nil?

          instance_variable_set "@#{name}".to_sym, relation
        end
      end
      def macro; :has_many; end
    end

    class HasOne < SingularAssociation
      def initialize(klass, name, options = {})
        super
        # Set the Getter:
        klass.send :define_method, name do
          value = self.instance_variable_get "@#{name}".to_sym
          if value.nil?
            reflection = self.class.reflect_on_association name
            foreign_key = (self.class.name.demodulize.underscore + '_id').to_sym
            value = self.instance_variable_set(
              "@#{name}".to_sym,
              reflection.associated_class.where(foreign_key => self.id).first
            )
          end
          value
        end
      end
      def macro; :has_one; end
    end

    class BelongsTo < SingularAssociation
      def initialize(klass, name, options = {})
        super

        # Define the foreign key:
        klass.send :field, foreign_key, foreign_key: true

        # Define the setter
        klass.send :define_method, "#{name}=".to_sym do |new_value|
          reflection = self.class.reflect_on_association(name)
          foreign_key_name = reflection.foreign_key
          
          new_value = reflection.associated_class.new new_value if new_value.is_a? Hash
          raise "Type Mismatch on #{self.class.name} ##{name}= #{new_value.inspect}" unless new_value.nil? || new_value.class == reflection.associated_class

          self.send "#{foreign_key_name}=".to_sym, new_value.nil? ? nil : new_value.id
          instance_variable_set "@#{name}".to_sym, new_value
        end
      end
      def macro; :belongs_to; end
      def foreign_key
        @foreign_key ||= options[:foreign_key] || "#{name}_id".to_sym
      end
    end

    def is_association?(association_name)
      self.class.is_association? association_name
    end

    module ClassMethods

      def is_association?(association_name)
        associations.has_key? association_name
      end

      def has_many(name, options={})
        associations[name] = HasMany.new self, name, options
      end

      def has_one(name, options={})
        associations[name] = HasOne.new self, name, options
      end

      def belongs_to(name, options={})
        associations[name] = BelongsTo.new self, name, options
      end

      def reflect_on_association(association_name)
        raise "Association #{name} does not exist" unless is_association? association_name
        associations[association_name]
      end

      def reflect_on_all_associations
        associations
      end

      private

        def associations
          @associations ||= {}.with_indifferent_access
        end

    end
  end

end
