require 'cgi'

module NaranyaEcm::Rest
  # = REST Errors
  #
  # Generic REST exception class.
  class RestError < StandardError

    attr_reader :response

    def initialize(given_response, msg = "")
      @response = given_response

      # Generate a default message if we got a blank message:
      if msg.blank?
        response_env = given_response.env
        msg = response_env.method.upcase.to_s
        msg += " '#{response_env.url.to_s}'"
        msg += " failed (HTTP #{response_env.status})"
      end

      # Record custom attributes in NewRelic:
      if ::NContent::SDK.config.new_relic_agent_available?
        ncontent_url = given_response.env.url

        # Parsear el query string a hash:
        ncontent_request_params = CGI.parse ncontent_url.query.to_s

        # Agregar al hash de ncontent_request_params el request_body como hash:
        if given_response.env.request_headers['Content-Type'] == 'application/json' && given_response.env[:request_body].present?
          ncontent_request_params.merge! ActiveSupport::JSON.decode(given_response.env[:request_body])
        end

        ::NewRelic::Agent.add_custom_attributes \
          ncontent_url: ncontent_url.to_s.split('?').first,
          request_params: ncontent_request_params
      end

      super(msg)
    end

    ##
    # Captura un response y genera una excepción.
    def self.build_from_failed_response(response, message = "")
      klass = case response.status
      when 401 then AuthorizationRequired
      when 403 then Forbidden
      when 404 then ResourceNotFound
      when 422 then UnprocessableEntity
      else RemoteFailed
      end

      klass.new response
    end

    ##
    # Captura un response y avienta una excepción.
    def self.raise_by_failed_response(response, message = "")
      raise build_from_failed_response(response, message)
    end

  end

  # Raised when an object assigned to an association has an incorrect type.
  #
  #   class Ticket < ActiveRecord::Base
  #     has_many :patches
  #   end
  #
  #   class Patch < ActiveRecord::Base
  #     belongs_to :ticket
  #   end
  #
  #   # Comments are not patches, this assignment raises AssociationTypeMismatch.
  #   @ticket.patches << Comment.new(content: "Please attach tests to your patch.")
  class AssociationTypeMismatch < RestError
  end

  # Raised when unserialized object's type mismatches one specified for serializable field.
  class SerializationTypeMismatch < RestError
  end

  # Raised when adapter not specified on connection (or configuration file <tt>config/database.yml</tt>
  # misses adapter field).
  class AdapterNotSpecified < RestError
  end

  # Raised when REST cannot find database adapter specified in <tt>config/database.yml</tt> or programmatically.
  class AdapterNotFound < RestError
  end

  # Raised when connection to the database could not been established (for example when <tt>connection=</tt>
  # is given a nil object).
  class ConnectionNotEstablished < RestError
  end

  # Raised when REST cannot find record by given id or set of ids.
  class ResourceNotFound < RestError
  end

  # Raised by ActiveRecord::Base.save! and ActiveRecord::Base.create! methods when record cannot be
  # saved because record is invalid.
  class ResourceNotSaved < RestError
  end

  # Raised by ActiveRecord::Base.destroy! when a call to destroy would return false.
  class ResourceNotDestroyed < RestError
  end

  class AuthorizationRequired < RestError

    def self.parse_www_authenticate_header(response)
      Hash[response.headers['WWW-Authenticate'].scan(/(\w+)="([^"]+)"/)] \
        if response.headers['WWW-Authenticate']
    end

    delegate :parse_www_authenticate_header, to: :class

    def initialize(given_response, msg = "")
      data = parse_www_authenticate_header given_response
      super given_response,
        "#{data['error_description']} (HTTP status #{given_response.status})"
    end
  end

  class Forbidden < RestError
  end

  class UnprocessableEntity < RestError
  end

  class RemoteFailed < RestError
  end

end
