require 'active_model'

require 'naranya_ecm/rest/notifications'

module NaranyaEcm::Rest
  module Model

    # Constant proc that validates if a given value is a boolean:
    BOOLEAN = Proc.new { |value| !!value == value }

    class LocalizedAttribute < HashWithIndifferentAccess

      alias_method :locales, :keys
      alias_method :available_locales, :locales

      def initialize(*args)
        super()
        given_attrs = args.shift
        given_attrs.each { |k,v| self[k] = v } unless given_attrs.nil?
      end

      def lookup
        locale = ::I18n.locale
        if ::I18n.respond_to?(:fallbacks)
          lookup_result = self[::I18n.fallbacks[locale].map(&:to_s).find{ |loc| self.has_key?(loc) }]
        else
          lookup_result = self[locale.to_s]
        end
        lookup_result = self.present? ? self.first[1] : nil unless lookup_result.present?
        lookup_result
      end

      def method_missing(method_name, *args, &block)
        if method_name == self.class.aliased_lookup_name
          # El nombre del método debe ser un alias de lookup.
          # Agregar el alias:
          self.class.send :alias_method, self.class.aliased_lookup_name, :lookup
          # llamar el método de lookup. La siguiente ocación no caerá aqui:
          lookup
        else
          super # You *must* call super if you don't handle the
          # method, otherwise you'll mess up Ruby's method
          # lookup.
        end
      end

      class << self
        def aliased_lookup_name
          @lookup_alias ||= self.name.demodulize[0..-("Translations".length+1)].underscore.to_sym
        end
      end
    end

    extend ActiveSupport::Concern

    included do

      include ActiveModel::Model
      extend  ActiveModel::Callbacks

      include ActiveModel::Dirty
      primary_key :id, type: String

      # Ignore the server resource's url attribute:
      attr_accessor :url

      include NaranyaEcm::Rest::Notifications

      include NaranyaEcm::Rest::Associations
      include NaranyaEcm::Rest::Persistence

      extend NaranyaEcm::Rest::FinderMethods

      alias_method_chain :initialize, :defaults

      if NContent::SDK.config.extra_attributes.present? and
      extra_fields = NContent::SDK.config.extra_attributes[self.name.demodulize.underscore]

        extra_fields.each do |name, type|

          field_options = {}

          field_options[:type] = case type
          when 'boolean'
            NaranyaEcm::Rest::Model::BOOLEAN
          else
            type.classify.constantize
          end

          field(name.to_sym, field_options)
        end
      end

    end

    def initialize_with_defaults(*args)
      init_internals

      initialize_without_defaults(*args)
      # inicializar con los valores default los campos que esten nulos y tengan un default definido:
      self.class.fields.values
        .select { |f| instance_variable_get("@#{f[:name]}".to_sym).nil? && f[:default] }
        .each { |f| instance_variable_set("@#{f[:name]}".to_sym, f[:default].call) }

    end

    def method_missing(method, *args, &block)
      if method =~ /^(\w+)=$/i
        self.class.send :field, $1.to_sym, type: args.first.class
        self.send method, *args
      else
        super # You *must* call super if you don't handle the
              # method, otherwise you'll mess up Ruby's method
              # lookup.
      end
    end

    def attributes
      self.class.fields.values.inject('id' => self.id) do |hsh, attr_data|
        hsh[attr_data[:name].to_s] = self.public_send attr_data[:name]
        hsh
      end.with_indifferent_access
    end

    # Returns an <tt>#inspect</tt>-like string for the value of the
    # attribute +attr_name+. String attributes are truncated upto 50
    # characters, Date and Time attributes are returned in the
    # <tt>:db</tt> format, Array attributes are truncated upto 10 values.
    # Other attributes return the value of <tt>#inspect</tt> without
    # modification.
    #
    #   person = Person.create!(name: 'David Heinemeier Hansson ' * 3)
    #
    #   person.attribute_for_inspect(:name)
    #   # => "\"David Heinemeier Hansson David Heinemeier Hansson ...\""
    #
    #   person.attribute_for_inspect(:created_at)
    #   # => "\"2012-10-22 00:15:07\""
    #
    #   person.attribute_for_inspect(:tag_ids)
    #   # => "[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ...]"
    def attribute_for_inspect(attr_name)
      value = self.public_send(attr_name)

      if value.is_a?(String) && value.length > 50
        "#{value[0, 50]}...".inspect
      elsif value.is_a?(Date) || value.is_a?(Time)
        %("#{value.to_s(:db)}")
      elsif value.is_a?(Array) && value.size > 10
        inspected = value.first(10).inspect
        %(#{inspected[0...-1]}, ...])
      else
        value.inspect
      end
    end

    def associations
      self.instance_variables
        .map { |instance_variable| instance_variable[1..-1].to_sym }
        .select { |accessor_name| self.is_association?(accessor_name) }
        .inject({}.with_indifferent_access) { |hash, accessor_name| hash[accessor_name] = self.send accessor_name; hash }
    end

    def path
      "#{self.class.path}/#{self.id}"
    end

    # Returns the contents of the record as a nicely formatted string.
    def inspect
      # We check defined?(@attributes) not to issue warnings if the object is
      # allocated but not initialized.
      inspection = (['id'] + self.class.fields.keys.sort).map { |name|
        value = attribute_for_inspect(name)
        "#{name}: #{value}" unless value == 'nil'
      }.compact.join(", ")

      inspection = "not initialized" if inspection.blank?

      "#<#{self.class} #{inspection}>"
    end

    module ClassMethods

      def fields
        @_fields ||= {}.with_indifferent_access
      end

      def primary_key(name, options = {})
        attr_accessor name
        define_attribute_method name
      end

      def field(name, options = {})

        if options[:localize]
          lookup_alias_name = name
          name = "#{name}_translations".to_sym
        end

        fields[name] = {name: name}.merge(options).with_indifferent_access

        attr_accessor name
        define_attribute_method name

        if options[:localize]
          # Define localized setter with dirty notification:
          define_method("#{name}=".to_sym) do |value|
            value = {} if value.nil?
            value = LocalizedAttribute.new value if value.is_a?(Hash) && !value.is_a?(LocalizedAttribute)

            raise "Type Mismatch..." unless value.is_a?(LocalizedAttribute)

            variable_name = "@#{name}".to_sym
            self.send "#{name}_will_change!".to_sym unless value == instance_variable_get(variable_name)
            instance_variable_set variable_name, value
          end

          # Define lookup alias:
          define_method lookup_alias_name do
            self.send(name).lookup
          end

        else
          # Define normal setter with dirty notification:
          define_method("#{name}=".to_sym) do |value|
            variable_name = "@#{name}".to_sym
            self.send "#{name}_will_change!".to_sym unless value == instance_variable_get(variable_name)
            instance_variable_set variable_name, value
          end
        end
      end

      def path
        @path ||= "/#{self.name.demodulize.tableize}"
      end

      def where(given_conditions={})
        Relation.new self, given_conditions.with_indifferent_access
      end

      delegate :first, :last, to: :all

      def all
        where()
      end

      protected

        def _load_from_fetched_attributes(*args)
          loaded = self.new
          loaded.send :_load_from_fetched_attributes!, *args
          loaded
        end

    end

    # Returns +true+ if the resource is read only.
    def readonly?
      @readonly
    end

    # Marks this record as read only.
    def readonly!
      @readonly = true
    end

    protected

      def off_band_changes
        @off_band_changes ||= {}.with_indifferent_access
      end

    private

      def _load_from_fetched_attributes!(params={})
        if params

          association_dictionary = self.class.send(:associations).inject({}) do |hash, keyval|
            key, val = keyval
            hash[key] = val.associated_class
            hash
          end

          nested_attributes = params.slice *association_dictionary.keys
          self_attributes = params.reject { |k,v| nested_attributes.keys.include? k }

          loaded_associations = nested_attributes.inject({}) do |hash, keyval|
            association_name, association_attributes = keyval
            association_class = association_dictionary[association_name]

            hash[association_name] = if association_attributes.is_a?(Array)
              association_attributes.map do |x|
                association_class.send :_load_from_fetched_attributes, x
              end
            else
              association_class.send :_load_from_fetched_attributes, association_attributes
            end
            hash
          end

          # Set the given params and the inflated association params:
          self_attributes.merge(loaded_associations).each do |attr, value|
            self.public_send("#{attr}=", value)
          end

          changed_attributes.clear
        end
        @new_resource = false
      end

      def init_internals
        @readonly                 = false
        @destroyed                = false
        @new_resource             = true
      end

  end
end
