module NaranyaEcm::Rest

  module Persistence
    extend ActiveSupport::Concern

    class ReadOnlyResource < StandardError
    end

    class ResourceNotSaved < StandardError
    end

    class ResourceNotDestroyed < StandardError
    end

    module ClassMethods

      # Updates an object (or multiple objects) and saves it to the server, if
      # validations pass.
      #
      # The resulting object is returned whether the object was saved
      # successfully to the server or not.
      #
      # ==== Parameters
      #
      # * +id+ - This should be the id or an array of ids to be updated.
      # * +attributes+ - This should be a hash of attributes or an array of hashes.
      #
      # ==== Examples
      #
      #   # Updates one resource
      #   Person.update(15, user_name: 'Samuel', group: 'expert')
      #
      #   # Updates multiple resources
      #   people = { 1 => { "first_name" => "David" }, 2 => { "first_name" => "Jeremy" } }
      #   Person.update(people.keys, people.values)
      def update(id, attributes)
        if id.is_a?(Array)
          id.map.with_index { |one_id, idx| update(one_id, attributes[idx]) }
        else
          object = find(id)
          object.update(attributes)
          object
        end
      end

      def create(attributes={})
        object_to_create = self.new attributes
        object_to_create.save
        object_to_create
      end

      def create!(attributes={})
        created_object = create(attributes)
        raise "ValidationError" unless created_object.persisted?
        created_object
      end

      def delete(id, options={})
        ActiveSupport::Notifications.instrument('ncontent_sdk.destroy') do |notification_payload|
          response = NContent::SDK::RESTClient.delete("#{self.path}/#{id}.json")
          notification_payload.merge!({ id: id }.merge(self.as_notification_payload))
        end
        true
      end

    end

    included do
      define_model_callbacks :validation, :create, :update, :save, :destroy
    end

    def touch
      @is_touch = true
      self.update({})
    end

    # Returns true if this object hasn't been saved yet -- that is, a resource
    # for the object doesn't exist at the server yet; otherwise, returns false.
    def new_resource?
      @new_resource
    end

    alias_method :new?, :new_resource?
    alias_method :new_record?, :new_resource?

    # Returns true if this object has been destroyed, otherwise returns false.
    def destroyed?
      @destroyed
    end

    # Returns true if the resource is persisted, i.e. it's not a new resource
    # and it was not destroyed, otherwise returns false.
    def persisted?
      !(new_resource? || destroyed?)
    end

    # Saves the resource at the server.
    #
    # If the model is new a resource gets created at the server, otherwise
    # the existing resource gets updated.
    #
    # By default, save always run validations. If any of them fail the action
    # is cancelled and +save+ returns +false+. However, if you supply
    # validate: false, validations are bypassed altogether. See
    # ActiveModel::Validations for more information.
    #
    # There's a series of callbacks associated with +save+. If any of the
    # <tt>before_*</tt> callbacks return +false+ the action is cancelled and
    # +save+ returns +false+. See ActiveModel::Callbacks for further
    # details.
    #
    # Attributes marked as readonly are silently ignored if the resource is
    # being updated.
    def save(*)
      run_callbacks :save do
        @previously_changed = changes
        result = create_or_update
        @changed_attributes.clear
        result
      end
    end

    # Saves the model.
    #
    # If the model is new a resource gets created at the server, otherwise
    # the existing resource gets updated.
    #
    # With <tt>save!</tt> validations always run. If any of them fail
    # ActiveModel::resourceInvalid gets raised. See ActiveModel::Validations
    # for more information.
    #
    # There's a series of callbacks associated with <tt>save!</tt>. If any of
    # the <tt>before_*</tt> callbacks return +false+ the action is cancelled
    # and <tt>save!</tt> raises ActiveModel::resourceNotSaved. See
    # ActiveModel::Callbacks for further details.
    #
    # Attributes marked as readonly are silently ignored if the resource is
    # being updated.
    def save!(*)
      create_or_update || raise(resourceNotSaved.new(nil, self))
    end

    # Deletes the resource at the server and freezes this instance to
    # reflect that no changes should be made (since they can't be
    # persisted). Returns the frozen instance.
    #
    # The row is simply removed with an SQL +DELETE+ statement on the
    # resource's primary key, and no callbacks are executed.
    #
    # To enforce the object's +before_destroy+ and +after_destroy+
    # callbacks or any <tt>:dependent</tt> association
    # options, use <tt>#destroy</tt>.
    def delete
      self.class.delete(id) if persisted?
      @destroyed = true
      freeze
    end

    # Deletes the resource at the server and freezes this instance to reflect
    # that no changes should be made (since they can't be persisted).
    #
    # There's a series of callbacks associated with <tt>destroy</tt>. If
    # the <tt>before_destroy</tt> callback return +false+ the action is cancelled
    # and <tt>destroy</tt> returns +false+. See
    # Activeresource::Callbacks for further details.
    def destroy
      run_callbacks :destroy do
        #raise ReadOnlyresource if readonly?
        #destroy_associations
        self.class.delete(self.id) if persisted?
        @destroyed = true
        freeze
      end
    end

    # Deletes the resource at the server and freezes this instance to reflect
    # that no changes should be made (since they can't be persisted).
    #
    # There's a series of callbacks associated with <tt>destroy!</tt>. If
    # the <tt>before_destroy</tt> callback return +false+ the action is cancelled
    # and <tt>destroy!</tt> raises Activeresource::resourceNotDestroyed. See
    # Activeresource::Callbacks for further details.
    def destroy!
      destroy || raise(ResourceNotDestroyed, self)
    end

    # Updates a single attribute and saves the resource.
    # This is especially useful for boolean flags on existing resources.
    # Also note that
    #
    # * Validation is skipped.
    # * Callbacks are invoked.
    # * updated_at/updated_on attribute is updated if that attribute is available.
    # * Updates all the attributes that are dirty in this object.
    #
    # This method raises an +Activeresource::ActiveresourceError+  if the
    # attribute is marked as readonly.
    #
    # See also +update_column+.
    def update_attribute(name, value)
      name = name.to_s
      #verify_readonly_attribute(name)
      send("#{name}=", value)
      save(validate: false)
    end

    # Updates the attributes of the model from the passed-in hash and saves the
    # resource, all wrapped in a transaction. If the object is invalid, the
    # saving will fail and false will be returned.
    def update(attributes)
      assign_attributes(attributes)
      save
    end

    def assign_attributes(given_attributes)
      given_attributes.each do |key, val|
        self.public_send "#{key}=", val
      end
    end

    alias update_attributes update

    # Updates its receiver just like +update+ but calls <tt>save!</tt> instead
    # of +save+, so an exception is raised if the resource is invalid.
    def update!(attributes)
      # The following transaction covers any possible server side-effects of the
      # attributes assignment. For example, setting the IDs of a child collection.
      with_transaction_returning_status do
        assign_attributes(attributes)
        save!
      end
    end

    alias update_attributes! update!

    protected

      def create_or_update
        raise ReadOnlyResource, "#{self.class} is marked as readonly" if readonly?
        result = new_resource? ? _create_resource : _update_resource
        result != false
      end

      def load_server_errors(server_error_hash)
        server_error_hash = server_error_hash.with_indifferent_access
        if server_error_hash.present? && server_error_hash.has_key?(:errors) && server_error_hash[:errors].present?
          server_error_hash[:errors].each do |field_name, error_list|
            error_list.each do |message|
              self.errors.add field_name.to_sym, message
            end
          end
        else
          self.errors.add :base, 'is invalid in server'
        end
      end

      def reload
        response = NContent::SDK::RESTClient.get "#{path}.json"
        send :_load_from_fetched_attributes!, response.body
        true
      end

      def reload!; raise "Ja!" unless reload; end

      def persist_request_options(data)
        {
          body: ActiveSupport::JSON.encode({ self.class.name.demodulize.underscore => data }),
          headers: { 'Content-Type' => 'application/json'}
        }
      end

      def _create_resource
        run_callbacks :create do

          create_path = "#{self.class.path}.json"

          begin
            ActiveSupport::Notifications.instrument('ncontent_sdk.create') do |notification_payload|
              response = NContent::SDK::RESTClient.post create_path do |req|
                req.headers[:content_type] = 'application/json'

                # Obtener solo los cambios con respecto al objeto recién
                # inicializado:
                attributes_to_post = changes.inject({}) do |attr_hash, attribute_change|
                  attribute_name, attribute_change_values = attribute_change
                  previous_attribute_value, new_attribute_value = attribute_change_values
                  attr_hash[attribute_name] = new_attribute_value
                  attr_hash
                end

                req.body = ActiveSupport::JSON.encode(
                  self.class.name.demodulize.underscore => attributes_to_post
                )
              end

              send :_load_from_fetched_attributes!, response.body

              notification_payload.merge! self.as_notification_payload

            end
            true
          rescue UnprocessableEntity => e
            self.load_server_errors(e.response.body)
            false
          end
        end
      end

      def _update_resource
        run_callbacks :update do

          unless @is_touch
            attributes_to_save = self.changes.dup
            attributes_to_save.each do |key, val|
              attributes_to_save[key] = val.last
            end
          else
            # Encontrar el primer atributo con valor que no sea ID o lifecycle_state:
            key, val = self.attributes.detect do |k,v|
              !%w(id lifecycle_state).include?(k) && v.present?
            end
            attributes_to_save = { key => val }
          end

          attributes_to_save.merge! off_band_changes

          update_path = "#{self.path}.json"

          begin
            ActiveSupport::Notifications.instrument('ncontent_sdk.update') do |notification_payload|
              response = NContent::SDK::RESTClient.put update_path do |req|
                req.headers[:content_type] = 'application/json'
                req.body = ActiveSupport::JSON.encode(
                  self.class.name.demodulize.underscore => attributes_to_save
                )
              end

              self.reload!
              off_band_changes.clear if off_band_changes.present?
              notification_payload.merge!(self.as_notification_payload)
            end
            true
          rescue UnprocessableEntity
            self.load_server_errors(response.body)
            false
          end
        end
      end

  end

end
