require 'naranya_ecm/rest/relation'

module NaranyaEcm::Search
  class Results < NaranyaEcm::Rest::Relation

    DEFAULT_MATERIALIZER = Proc.new do |search_results|

      # Duplicate the hits array - we'll replace each element with the
      # resource:
      result_elements = search_results.hits.dup

      # Load from server all elements yet to materialize:
      pending_hits = result_elements.select { |e| e.is_a? NaranyaEcm::Search::Hit }
      pending_hits.map(&:klass).uniq.each do |hit_klass|
        # Obtain the ids for this hit_klass:
        hit_ids = pending_hits.select { |e| e.klass == hit_klass }.map(&:id)

        # Fetch the data from server:
        hit_klass.where(id: hit_ids).fetch_from_server.each do |fetched_data|
          r = hit_klass.load fetched_data

          hit_data = result_elements.detect { |e| e.is_a?(NaranyaEcm::Search::Hit) && e.klass == r.class && e.id == r.id }
          hit_index = result_elements.find_index hit_data

          r.merge_search_hit_data hit_data
          result_elements[hit_index] = r
        end
      end

      result_elements
    end

    delegate :get, to: NContent::SDK::RESTClient

    attr_accessor :materializer

    def total_hits
      do_search! unless @total_hits
      @total_hits
    end
    alias_method :total_count, :total_hits

    def digest
      [
        'naranya_ecm',
        'search',
        'results',
        Digest::MD5.hexdigest(@dsl.to_s)
      ]
    end

    def hits
      do_search! unless @hits
      @hits
    end

    def count
      hits.count
    end

    def any?
      count > 0
    end

    def empty?
      count < 1
    end

    def facts
      do_search! unless @facts
      @facts
    end

    def max_score
      do_search! unless @max_score
      @max_score
    end

    def query_process_time
      do_search! unless @query_process_time
      @query_process_time
    end

    def query_total_time
      do_search! unless @query_total_time
      @query_total_time
    end

    def materializer
      @materializer || DEFAULT_MATERIALIZER
    end

    def to_a
      materialize! unless @elements
      elements
    end

    private

      def do_search!

        start_time = Time.now

        # Strip '.json' from collection path:
        search_path = klass.present? ? klass.path : ''
        search_path += '/_search'

        search_response = get search_path do |req|
          req.body = ActiveSupport::JSON.encode(conditions)
        end

        # puts "curl -XGET '#{NContent::SDK.config.api_host}#{search_path}?pretty=true' -d '#{search_body}'"

        # Parse the search results:
        @facts = search_response.body.with_indifferent_access

        # Obtain the query processing time, which is given in miliseconds as 'took' key:
        @query_process_time = (Float(@facts.delete(:took)) / 1000)

        result_hits = @facts.delete :hits

        @total_hits = result_hits.delete :total
        @max_score  = result_hits.delete :max_score

        @hits = result_hits[:hits].each_with_index.map { |ht, o| Hit.new(ht.merge(order: o)) }

        @hits.freeze
        @facts.freeze

        @query_total_time = (Time.now - start_time)

      end

      def materialize!
        # Call the materializer and:
        # - Filter out elements that materialized as nulls
        # - Order the elements by relevance
        @elements = materializer.call(self)
          .select { |materialized_element| materialized_element.present? }
          .sort_by do |element|
            element.search_order
          end

      end

  end
end
