require 'oauth2'
require 'ncontent/sdk/faraday_middleware'

module NContent

  module SDK

    class RESTClient

      CLIENT_TOKEN_CACHE_KEY = "ncontent/api_client/client_token".freeze

      class << self
        delegate :config, :logger, :cache, to: NContent::SDK
      end

      # Make a request to the API:
      #
      # @param [Symbol] verb the HTTP request method
      # @param [String] path the HTTP URL path of the request
      # @param [Hash] opts the options to make the request with
      def self.request(verb, path=nil, opts = {}, &block)

        retrying = false

        begin

          # If we beforehand know that the current token expired then the client
          # token should be re-generated.
          reset_oauth_token! if oauth_token.expired?

          oauth_token.request(verb, path, opts) do |req|
            req.options.timeout = config.api_call_timeout
            req.options.open_timeout = config.api_call_open_timeout
            yield req if block_given?
          end

        rescue OAuth2::Error => error
          if error.response.status == 401 && !retrying
            reset_oauth_token!
            retrying = true
            logger.info "Retrying response #{verb.upcase} #{path} with a fresh nContent token."
            retry
          else
            logger.warn "Not retrying response #{verb.upcase} #{path}."

            # Extract from the OAuth::Error the base Faraday response:
            error_response = error.response.response

            # Build the error that will be raised:
            error_to_raise = NaranyaEcm::Rest::RestError.build_from_failed_response(error_response)

            raise error_to_raise
          end
        end
      end

      def self.get(path=nil, opts = {}, &block)
        request(:get, path, opts, &block)
      end

      def self.post(path=nil, opts = {}, &block)
        request(:post, path, opts, &block)
      end

      def self.put(path=nil, opts = {}, &block)
        request(:put, path, opts, &block)
      end

      def self.patch(path=nil, opts = {}, &block)
        request(:patch, path, opts, &block)
      end

      def self.delete(path=nil, opts = {}, &block)
        request(:delete, path, opts, &block)
      end

      private

        # The oauth client is accessed as a class method, but will exist one instance
        # per thread (hence the @ instead of @@ and using ||= without remorse)
        def self.oauth_client
          @oauth_client ||= OAuth2::Client.new(config.api_key, config.api_secret, site: config.api_host) do |faraday|
            ### Oauth2 Client (Faraday) builder:

            faraday.request :url_encoded # them posts...

            faraday.response :logger, logger

            faraday.use NContent::SDK::FaradayMiddleware::RequestBodyOnEnvironment

            faraday.use NContent::SDK::FaradayMiddleware::RESTAPICallBenchmark

            faraday.use NContent::SDK::FaradayMiddleware::ResponseParser

            # Require the typhoeus' faraday adapter if applies:
            require 'typhoeus/adapters/faraday' if config.api_adapter == 'typhoeus' && defined?(Typhoeus)

            faraday.adapter config.api_adapter.to_sym
          end
        end

        def self.oauth_token(overwrite_cache = false)
          @oauth_token ||= if !overwrite_cache && (cached_data = cache.read CLIENT_TOKEN_CACHE_KEY)
            logger.debug "Generating client token from cached data: #{cached_data.inspect}"
            OAuth2::AccessToken.from_hash oauth_client, cached_data
          else
            logger.info "Requesting new client token from nContent server"
            new_token = oauth_client.client_credentials.get_token
            cache.write CLIENT_TOKEN_CACHE_KEY, new_token.to_hash.with_indifferent_access
            new_token
          end
        end

        def self.reset_oauth_token!
          # if (cached_data = cache.read CLIENT_TOKEN_CACHE_KEY) && (cached_data['access_token'] == oauth_token.token)
            # The cached data matches the current oauth_token.
            # We'll unset @oauth_token, and call 'oauth_token' with 'true' to
            # force a cache overwrite:
            @oauth_token = nil
            oauth_token(true)
          # else
          #   # The cached data does not match with the current oauth_token.
          #   # Maybe the cached data was updated by another thread or process.
          #   @oauth_token = OAuth2::AccessToken.from_hash oauth_client, cached_data
          # end
        end

    end

  end

end
