require 'sinatra/base'
require 'bson'

module NContent
  module SDK
    module Testing

      ##
      # Clase que simula el servicio de nContent:
      # Ver https://robots.thoughtbot.com/how-to-stub-external-services-in-tests
      class ServerMock < Sinatra::Base

        class ValidationError < StandardError
          attr_reader :errors
          def initialize(errors = {})
            @errors = errors.collect({}) do |hsh, keyval|
              key, val = keyval
              hsh[key] = [] unless hsh[key].is_a? Array
              hsh[key] << val
              hsh
            end
          end
        end

        delegate :config, to: :class

        class << self
          delegate :config, to: NContent::SDK
        end

        COLLECTION_NAMES = %w(
          categories content_versions contents download_authorizations
          media_processes media_profiles media_resources
        ).freeze

        RESOURCE_TYPES = COLLECTION_NAMES.map(&:singularize).freeze

        def self.get_blank_collection_set
          COLLECTION_NAMES.inject({}) do |hash, collection_name|
            hash[collection_name] = {}
            hash
          end.with_indifferent_access
        end

        def self.clear_collections!
          @collections = nil
        end

        def self.collections
          @collections ||= get_blank_collection_set
        end

        def self.method_missing(method_name, *args, &block)
          if COLLECTION_NAMES.include? method_name.to_s

            # Create the method:
            define_singleton_method method_name do
              collections[method_name.to_s]
            end

            # Call the newly defined method:
            self.send method_name, *args
          elsif method_name =~ /\Aget_(\w+)\z/ && RESOURCE_TYPES.include?($1)

            # Create the method:
            define_singleton_method method_name do |resource_id|
              get_resource $1, resource_id
            end

            # Call the newly defined method:
            self.send method_name, *args
          elsif method_name =~ /\A(create|update|upsert)_(\w+)\z/ && RESOURCE_TYPES.include?($2)

            # Create the methods:
            %w(create update upsert).each do |method_prefix|
              new_method_name = "#{method_prefix}_#{$2}".to_sym
              define_singleton_method new_method_name do |*args|
                upsert_resource $2.pluralize, *args
              end
            end

            # Call the newly defined method:
            self.send method_name, *args
          else
            super # You *must* call super if you don't handle the
                  # method, otherwise you'll mess up Ruby's method
                  # lookup.
          end
        end

        def self.upsert_resource(collection_name, given_resource_attributes, resource_id = nil)
          # Ignore the ID:
          given_resource_attributes.delete 'id'

          invalid_attributes = given_resource_attributes.select do |k, v|
            v =~ /invalid/i
          end

          raise ValidationError, invalid_attributes.collect {} do |h, kv|
            k, v = kv;
            h[k] = ['is invalid']
            h
          end if invalid_attributes.any?

          if resource_id.present? && collections[collection_name].key?(resource_id)
            collections[collection_name][resource_id].merge! given_resource_attributes
          else
            # Caso especial:
            if collection_name == 'contents'
              given_resource_attributes.reverse_merge "lifecycle_state" => 'draft'

              # Generar iconos y screenshots en caso de que sea Game:
              if given_resource_attributes['type'] == "Game"
                icon = self.upsert_media_resource(
                  roles: ["icon"],
                  downloadable_url: "http://test-bucket.s3.amazonaws.com/icons/original/#{resource_id}.jpg",
                  type: "image/jpg",
                  access_type: "public",
                  parent_document_id: resource_id,
                  parent_document_type: "Content"
                )

                %w(120w 150w 300w).each do |processed_role|
                  self.upsert_media_resource(
                    roles: ["icon", "processed", "v2", processed_role],
                    downloadable_url: "http://test-bucket.s3.amazonaws.com/icons/#{processed_role}/#{resource_id}.jpg",
                    type: "image/jpg",
                    access_type: "public",
                    parent_document_id: icon[:id],
                    parent_document_type: "MediaResource"
                  )
                end

              end
            end

            resource_id = (BSON::ObjectId.new).to_s
            collections[collection_name][resource_id] = { id: resource_id }.merge(
              given_resource_attributes
            ).with_indifferent_access
          end

          collections[collection_name][resource_id]
        end

        # Éste método es útil para llamar "stub_request", usando el host+port configurado por ENV["NARANYA_ECM_SITE"]
        def self.url_matcher
          @url_matcher ||= begin
            site_url = URI config.api_host
            pattern_string = site_url.host
            pattern_string += ":#{site_url.port}" unless site_url.port == 80
            Regexp.new pattern_string, "i"
          end
        end

        def self.get_resource(resource_type, resource_id)
          collection_name = resource_type.pluralize
          requested_resource = self.collections[collection_name.to_s].fetch resource_id, nil

          # Traerse los MR's:
          requested_resource[:media_resources] = self.media_resources.values.select do |mr|
            mr[:parent_document_type] == "Content" && mr[:parent_document_id] == resource_id
          end.map do |mr|
            mr[:media_resources] = self.media_resources.values.select do |c_mr|
              c_mr[:parent_document_type] == "MediaResource" && c_mr[:parent_document_id] == mr[:id]
            end
            mr
          end if requested_resource.present? && collection_name == 'contents'

          requested_resource
        end

        def self.get_resource_headers(collection_name, resource_id)
          raise "no collection name given" unless collection_name.present?
          raise "no resource id given" unless resource_id.present?

          headers = {}
          headers = {
            "X-Frame-Options"         => "SAMEORIGIN",
            "X-Xss-Protection"        => "1; mode=block",
            "X-Content-Type-Options"  => "nosniff",
            "Location"                => "http://ecm.naranya.net:5000/#{collection_name}/#{resource_id}",
            "Content-Type"            => "application/json; charset=utf-8",
            "Vary"                    => "Accept-Encoding",
            "Etag"                    => '"a547f3ead872de01928f79724390aba4"',
            "Cache-Control"           => "max-age=0, private, must-revalidate",
            "Set-Cookie"              => "request_method=POST; path=/",
            "X-Request-Id"            => "a6463225-df54-450d-b571-7068f3c96364",
            "X-Runtime"               => "'0.241093'",
            "Transfer-Encoding"       => "chunked"
          } if self.collections[collection_name].key? resource_id
        end

        delegate  :collections, :get_resource, :get_resource_headers,
                  :upsert_resource, to: :class

        # Servicio de token de oauth:
        # Responderá con un token "fake" de prueba...
        post "/oauth/token" do
          case params[:grant_type]
          when "client_credentials"
            status 201

            headers \
              "X-Frame-Options" => "SAMEORIGIN",
              "X-XSS-Protection" => "1; mode=block",
              "X-Content-Type-Options" => "nosniff",
              "Cache-Control" => "no-store",
              "Pragma" => "no-cache",
              "Content-Type" => "application/json; charset=utf-8",
              "Vary" => "Accept-Encoding",
              "ETag" => '"75e236bc0e47293b0bfbb7a5d16a06c3"',
              "X-Request-Id" => "f4d38d78-84b9-4c41-8943-cf2c6f519061",
              "X-Runtime" => "0.015009",
              "Transfer-Encoding" => "chunked"

            body ActiveSupport::JSON.encode(
              access_token:  "NContentClientTestAccessTokenA",
              refresh_token: "NContentClientTestRefreshToken",
              token_type:    "bearer",
              expires_in:    7200
            )
          else
            status 401
          end

        end

        get '/:collection_name.json' do
          status 200

          headers \
            "X-Frame-Options"         => "SAMEORIGIN",
            "X-Xss-Protection"        => "1; mode=block",
            "X-Content-Type-Options"  => "nosniff",
            "Content-Type"            => "application/json; charset=utf-8",
            "Vary"                    => "Accept-Encoding",
            "Etag"                    => '"d751713988987e9331980363e24189ce"',
            "Cache-Control"           => "max-age=0, private, must-revalidate",
            "Set-Cookie"              => "request_method=GET; path=/",
            "X-Request-Id"            => "4fe9e39f-2b2d-400f-9da7-0b9ad611ce26",
            "X-Runtime"               => "'0.019181'",
            "Transfer-Encoding"       => "chunked"

          body ActiveSupport::JSON.encode collections[params[:collection_name]].values

        end

        post "/:collection_name.json" do
          collection_name = params[:collection_name]
          params = ActiveSupport::JSON.decode request.body.read
          attributes = params[collection_name.singularize]

          if attributes.present?
            begin
              resource = upsert_resource collection_name, attributes
              status 201
              headers get_resource_headers collection_name, resource[:id]
              body ActiveSupport::JSON.encode resource
            rescue ValidationError => e
              status 422
              body ActiveSupport::JSON.encode e.errors
            end
          else
            status 422
          end

        end

        get "/:collection_name/:id.json" do
          resource = get_resource params[:collection_name], params[:id]
          if resource.present?
            status 201
            headers get_resource_headers params[:collection_name], resource[:id]
            body ActiveSupport::JSON.encode resource
          else
            status 404
          end
        end

        put '/:collection_name/:id.json' do

          collection_name = params[:collection_name]
          resource_id     = params[:id]

          if collection_name != 'download_authorizations' and resource = get_resource(collection_name, resource_id)

            params = ActiveSupport::JSON.decode(request.body.read)

            if attributes = params[collection_name.singularize]
              begin
                resource = upsert_resource collection_name, attributes, resource_id
                status 200
                headers get_resource_headers collection_name, resource[:id]
                body ActiveSupport::JSON.encode resource
              rescue ValidationError => e
                status 422
                body ActiveSupport::JSON.encode e.errors
              end
            else
              status 422
            end

          else
            status 404
          end
        end

        get // do
          byebug
        end
      end

    end
  end
end
